from . test_amplitude_normalizer import TestAmplitudeNormalizer
from . test_bandpass_filter import TestBandpassFilter
from . test_base_processor import TestBaseProcessor
from . test_dc_offset_remover import TestDCOffsetRemover
from . test_end_frame_cutter import TestEndFrameCutter
from . test_full_wave_rectifier import TestFullWaveRectifier
from . test_linear_envelope import TestLinearEnvelope
from . test_segmenter import TestSegmenter

__all__ = [
    'TestAmplitudeNormalizer',
    'TestBandpassFilter',
    'TestBaseProcessor',
    'TestDCOffsetRemover',
    'TestEndFrameCutter',
    'TestFullWaveRectifier',
    'TestLinearEnvelope',
    'TestSegmenter',
]



from . base import BaseProcessor
import numpy as np


class AmplitudeNormalizer(BaseProcessor):
    """Amplitude normalizer for EMG signals

    """

    def __init__(self):
        pass

    def apply(self, x, **kwargs):
        """Apply amplitude normalizer

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels).
            Signal data to be processed.

        divisor : scalar, list, or ndarray
            One or more positive values.
            If x is in 1-dim or n_channels is 1, then divisor should be
            one value; otherwise divisor should be n_channels values.

        Returns
        -------
        x_processed : ndarray
            Same shape as x.
            The result of applying amplitude normalizer to x.
        """

        super().assert_input(x)

        assert 'divisor' in kwargs.keys(), 'Need to provide parameter "divisor" in AmplitudeNormalizer.apply'
        divisor = kwargs['divisor']
        divisor = np.array(divisor).flatten()
        if x.ndim == 1 or (x.ndim == 2 and x.shape[1] == 1):
            assert divisor.shape == (1,), 'divisor must be one number'
        else:
            assert divisor.shape == (x.shape[1],), 'divisor must be x.shape[1] numbers'

        assert np.all(divisor > 0), 'divisor must contain positive values'

        x_processed = x / divisor
        return x_processed

    def get_param_values_in_str(self):
        """Getting the parameter values of the processor for display
        purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = super().get_param_values_in_str()
        return params_in_str



from . base import BaseProcessor
import numpy as np


class DCOffsetRemover(BaseProcessor):
    """DC offset remover for EMG signals

    """

    def __init__(self):
        pass

    def apply(self, x, **kwargs):
        """Apply DC offset remover

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels).
            Signal data to be processed.

        Returns
        -------
        x_processed : ndarray
            Same shape as x.
            The result of applying DC offset remover to x.
        """

        super().assert_input(x)

        x_processed = x - np.mean(x, axis=0)
        return x_processed

    def get_param_values_in_str(self):
        """Getting the parameter values of the processor for display
        purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = super().get_param_values_in_str()
        return params_in_str



from . amplitude_normalizer import AmplitudeNormalizer
from . bandpass_filter import BandpassFilter
from . base import BaseProcessor
from . dc_offset_remover import DCOffsetRemover
from . end_frame_cutter import EndFrameCutter
from . full_wave_rectifier import FullWaveRectifier
from . linear_envelope import LinearEnvelope
from . segmenter import Segmenter

__all__ = [
    "AmplitudeNormalizer",
    "BandpassFilter",
    "BaseProcessor",
    "DCOffsetRemover",
    "EndFrameCutter",
    "FullWaveRectifier",
    "LinearEnvelope",
    "Segmenter",
]



from .. processors import *
from .. plots import *
import copy


class EMGMeasurement:
    """Wrapper of single-trial EMG processing

    Parameters
    ----------
    data : ndarray of shape (n_samples,) or (n_samples, n_channels),
        where n_samples > 15 (See Notes)
        Signal data of one trial.

    hz : float
        Sample rate in hertz.
        See class EMGMeasurementCollection for suggested values of hz.

    timestamp : ndarray or None, default None
        The actual timestamp corresponds to the signal. If it is an
        ndarray, it should be in 1-dim and have the same length as the
        first dimension of x.

    trial_name : str or None, default None
        Trial name.

    channel_names : list of str, or None, default None
        If list, its length should be equal to n_channels.
        Channel names to be shown in the plot.

    emg_plot_params : EMGPlotParams, default None
        Parameters to control the plot. See class EMGPlotParams and
        function emg_plot.

    Notes
    -----
    n_samples has to meet the length condition for using bandpass
    filter. With the default parameter in BandpassFilter, n_samples
    must exceed 15. (See class BandpassFilter, function apply)
    """

    def __init__(self, data, hz, timestamp=None, trial_name=None, channel_names=None, emg_plot_params=None):
        BaseProcessor.assert_input(data)
        assert data.shape[0] > 15, 'first dimension of x must have length > 15'
        self.data = copy.deepcopy(data)
        self.hz = hz
        self.timestamp = copy.deepcopy(timestamp)
        self.timestamp = BaseProcessor.get_timestamp(self.data.shape, self.timestamp, self.hz)
        self.trial_name = trial_name
        self.channel_names = channel_names
        self.emg_plot_params = emg_plot_params

    def apply_dc_offset_remover(self):
        """Apply DC offset remover to the data

        Returns
        -------
        None
        """

        self.data = DCOffsetRemover().apply(self.data)

    def apply_bandpass_filter(self, bf_order=4, bf_cutoff_fq_lo=10, bf_cutoff_fq_hi=450):
        """Apply bandpass filter to the data

        Parameters
        ----------
        bf_order : int, default=4
            Effective order (i.e., order after two-directional
            filtering) of the butterworth filter. bf_order should be
            a multiple of 2.

        bf_cutoff_fq_lo : float, default=10
            Low cutoff frequency of the bandpass filter.
            See class BandpassFilter.

        bf_cutoff_fq_hi : float, default=450
            High cutoff frequency of the bandpass filter.
            See class BandpassFilter.

        Returns
        -------
        None
        """

        self.data = BandpassFilter(self.hz, bf_order, bf_cutoff_fq_lo, bf_cutoff_fq_hi).apply(self.data)

    def apply_full_wave_rectifier(self):
        """Apply full wave rectifier to the data

        Returns
        -------
        None
        """

        self.data = FullWaveRectifier().apply(self.data)

    def apply_linear_envelope(self, le_order=4, le_cutoff_fq=6):
        """Apply linear envelope to the data

        Parameters
        ----------
        le_order : int, default=4
            Effective order (i.e., order after two-directional
            filtering) of the butterworth filter for linear envelope.
            le_order should be a multiple of 2.

        le_cutoff_fq : float, default=6
            Cutoff frequency of the lowpass filter.
            See class LinearEnvelope.

        Returns
        -------
        None
        """

        self.data = LinearEnvelope(self.hz, le_order, le_cutoff_fq).apply(self.data)

    def apply_end_frame_cutter(self, n_end_frames=30):
        """Apply end frame cutter to the data (signal and timestamp)

        Parameters
        ----------
        n_end_frames : int, default=30
            Number of frames to be cut off in both ends of the signal.
            n_end_frames >= 0.

        Returns
        -------
        None
        """

        self.data = EndFrameCutter(n_end_frames).apply(self.data)
        self.timestamp = EndFrameCutter(n_end_frames).apply(self.timestamp)

    def apply_amplitude_normalizer(self, max_amplitude):
        """Apply amplitude normalizer to the data

        Parameters
        ----------
        max_amplitude : scalar, list, or ndarray
            One or more positive values.
            If data is in 1-dim or n_channels is 1, then
            max_amplitude should be one value; otherwise max_amplitude
            should be n_channels values.
            max_amplitude is the value used as divisor in amplitude
            normalization.

        Returns
        -------
        None
        """

        self.data = AmplitudeNormalizer().apply(self.data, divisor=max_amplitude)

    def apply_segmenter(self, beg_ts, end_ts):
        """Apply segmenter to the data (signal and timestamp)

        Parameters
        ----------
        beg_ts : float
            Beginning time of interest. beg_ts <= timestamp[-1].

        end_ts : float
            End time of interest. end_ts >= timestamp[0].
            Also beg_ts < end_ts.

        Returns
        -------
        None
        """

        beg_idx, end_idx = BaseProcessor.get_indices_from_timestamp(self.timestamp, beg_ts, end_ts)
        self.data = Segmenter().apply(self.data, beg_idx=beg_idx, end_idx=end_idx)
        self.timestamp = Segmenter().apply(self.timestamp, beg_idx=beg_idx, end_idx=end_idx)

    def plot(self):
        """Plot the data

        Returns
        -------
        None
        """
        plot_emg(self.data, self.timestamp, channel_names=self.channel_names,
                 main_title=self.trial_name, emg_plot_params=self.emg_plot_params)

    def export_csv(self, csv_path):
        """Export the processing result to csv

        Parameters
        ----------
        csv_path : str
            The destination path to export data.

        Returns
        -------
        None
        """

        BaseProcessor.export_csv(csv_path, self.data, timestamp=self.timestamp, channel_names=self.channel_names)



from unittest import TestCase
import numpy as np
from pyemgpipeline.processors import BaseProcessor


class TestBaseProcessor(TestCase):

    def test_assert_input(self):
        # require: x be an ndarray
        with self.assertRaises(AssertionError):
            x = [20.3, 41.0, 53.9, 63.3]  # a list
            BaseProcessor.assert_input(x)

        # require: x be 1-dim or 2-dim ndarray
        with self.assertRaises(AssertionError):
            x = np.array(20.3)  # x.ndim = 0
            BaseProcessor.assert_input(x)

        # require: x be 1-dim or 2-dim ndarray
        with self.assertRaises(AssertionError):
            x = np.array([[[20.3], [41.0]], [[53.9], [63.3]], [[39.5], [24.9]]])  # x.ndim = 3
            BaseProcessor.assert_input(x)

        # require: x have no nan values
        with self.assertRaises(AssertionError):
            x = np.array([20.3, 41.0, np.nan, 63.3])
            BaseProcessor.assert_input(x)

    def test_get_timestamp(self):
        # case 1: output from timestamp
        x_shape = (5,)
        timestamp = np.array([0.2, 0.5, 0.6, 1.0, 1.2])
        hz = 10
        actual = BaseProcessor.get_timestamp(x_shape, timestamp=timestamp, hz=hz)
        desired = np.array([0.2, 0.5, 0.6, 1.0, 1.2])
        np.testing.assert_allclose(actual, desired)

        # case 2: output from hz
        x_shape = (5,)
        timestamp = None
        hz = 10
        actual = BaseProcessor.get_timestamp(x_shape, timestamp=timestamp, hz=hz)
        desired = np.array([0.0, 0.1, 0.2, 0.3, 0.4])
        np.testing.assert_allclose(actual, desired)

        # case 3: output from x_shape
        x_shape = (5,)
        timestamp = None
        hz = None
        actual = BaseProcessor.get_timestamp(x_shape, timestamp=timestamp, hz=hz)
        desired = np.array([0, 1, 2, 3, 4])
        np.testing.assert_allclose(actual, desired)

    def test_get_timestamp__assertion_raise(self):
        x_shape = (5,)  # x_shape[0] = 5
        hz = 10

        # require: timestamp be a 1-dim ndarray
        with self.assertRaises(AssertionError):
            timestamp = [0.2, 0.5, 0.6, 1.0, 1.2]  # a list
            BaseProcessor.get_timestamp(x_shape, timestamp=timestamp, hz=hz)

        # require: timestamp have length equal to x_shape[0]
        with self.assertRaises(AssertionError):
            timestamp = np.array([0.2, 0.5, 0.6, 1.0])  # len(timestamp) = 4
            BaseProcessor.get_timestamp(x_shape, timestamp=timestamp, hz=hz)

    def test_get_indices_from_timestamp(self):
        timestamp = np.array([0.2, 0.5, 0.6, 1.0, 1.2])

        beg_ts1 = 0.2
        end_ts1 = 1.1
        actual1 = BaseProcessor.get_indices_from_timestamp(timestamp, beg_ts=beg_ts1, end_ts=end_ts1)
        desired1 = (0, 3)
        self.assertEqual(actual1, desired1)

        beg_ts2 = 0
        end_ts2 = 1.2
        actual2 = BaseProcessor.get_indices_from_timestamp(timestamp, beg_ts=beg_ts2, end_ts=end_ts2)
        desired2 = (0, 4)
        self.assertEqual(actual2, desired2)

        beg_ts3 = 0.5
        end_ts3 = 2
        actual3 = BaseProcessor.get_indices_from_timestamp(timestamp, beg_ts=beg_ts3, end_ts=end_ts3)
        desired3 = (1, 4)
        self.assertEqual(actual3, desired3)

    def test_get_indices_from_timestamp__assertion_raise(self):
        timestamp = np.array([0.2, 0.5, 0.6, 1.0, 1.2])

        # require: beg_ts <= timestamp[-1]
        with self.assertRaises(AssertionError):
            beg_ts = 1.3
            end_ts = 2.0
            BaseProcessor.get_indices_from_timestamp(timestamp, beg_ts=beg_ts, end_ts=end_ts)

        # require: end_ts >= timestamp[0]
        with self.assertRaises(AssertionError):
            beg_ts = 0
            end_ts = 0.1
            BaseProcessor.get_indices_from_timestamp(timestamp, beg_ts=beg_ts, end_ts=end_ts)

        # require: beg_ts < end_ts
        with self.assertRaises(AssertionError):
            beg_ts = 0.7
            end_ts = 0.7
            BaseProcessor.get_indices_from_timestamp(timestamp, beg_ts=beg_ts, end_ts=end_ts)



from .. processors import *
from .. plots import *
from . emg_measurement_collection import EMGMeasurementCollection
import copy


class DataProcessingManager:
    """High-level, guided processing interface with accepted EMG
    processing conventions

    Parameters
    ----------
    c_raw : EMGMeasurementCollection or None
        Containing data (signal, timestamp, hz, channel names, etc.)
        and plot parameters.
        c_raw accepts data from method set_data_and_params and its
        value won't change when running method process_all.

    c : EMGMeasurementCollection or None
        Containing data (signal, timestamp, hz, channel names, etc.)
        and plot parameters.
        When running method process_all, c gets a fresh copy of raw
        data from c_raw. In this way method process_all can be
        repeatedly executed.

    dc_offset_remover : DCOffsetRemover or None
        A DCOffsetRemover processor.

    bandpass_filter : BandpassFilter or None
        A BandpassFilter processor.

    full_wave_rectifier : FullWaveRectifier or None
        A FullWaveRectifier processor.

    linear_envelope : LinearEnvelope or None
        A LinearEnvelope processor.

    end_frame_cutter : EndFrameCutter or None
        An EndFrameCutter processor.

    amplitude_normalizer : AmplitudeNormalizer or None
        An AmplitudeNormalizer processor.

    segmenter : Segmenter or None
        A Segmenter processor.

    segmenter_all_beg_ts : list or None
        Beginning time of interest for each trial to segment. See
        function apply_segmenter of class EMGMeasurementCollection.

    segmenter_all_end_ts : list or None
        End time of interest for each trial to segment. See
        function apply_segmenter of class EMGMeasurementCollection.
    """

    def __init__(self):
        self.c_raw = None
        self.c = None

        self.dc_offset_remover = None
        self.bandpass_filter = None
        self.full_wave_rectifier = None
        self.linear_envelope = None
        self.end_frame_cutter = None
        self.amplitude_normalizer = None
        self.segmenter = None

        self.segmenter_all_beg_ts = None
        self.segmenter_all_end_ts = None

    def set_data_and_params(self, all_data, hz, all_timestamp=None,
                            trial_names=None, channel_names=None, emg_plot_params=None):
        """Set the data and plot parameters and initialize default processors

        Parameters
        ----------
        all_data : list
            Elements of all_data are signal data of the trials.
            Signal data of each trial should be ndarray of shape
            (n_samples,) or (n_samples, n_channels), where n_samples > 15.
            Dimensions and n_channels (if 2-dim) of all trials should be
            the same.

        hz : float
            Sample rate in hertz.

        all_timestamp : list or None, default None
            If list, its length should be the same as the length of all_data
            and elements are ndarray or None.
            The ndarray in all_timestamp is the actual timestamp of the
            corresponding trial, and it should be in 1-dim and have the same
            length as the first dimension of its corresponding element in
            all_data.

        trial_names : list or None, default None
            Trial names.
            If not None, trial_names and all_data should have the same
            length.

        channel_names : list or None, default None
            If list, elements are str and its length should be equal to
            n_channels.
            Channel names of all trials to be shown in plots.

        emg_plot_params : EMGPlotParams or None, default None
            See class EMGPlotParams and function emg_plot.
        """

        self.c_raw = EMGMeasurementCollection(
            all_data, hz, all_timestamp=all_timestamp, trial_names=trial_names,
            channel_names=channel_names, emg_plot_params=emg_plot_params)

        # set the default 5 processors in case they are None
        if self.dc_offset_remover is None:
            self.dc_offset_remover = DCOffsetRemover()
        if self.bandpass_filter is None:
            self.bandpass_filter = BandpassFilter(self.c_raw.hz)
        if self.full_wave_rectifier is None:
            self.full_wave_rectifier = FullWaveRectifier()
        if self.linear_envelope is None:
            self.linear_envelope = LinearEnvelope(self.c_raw.hz)
        if self.end_frame_cutter is None:
            self.end_frame_cutter = EndFrameCutter()

    def set_dc_offset_remover(self, dc_offset_remover):
        """Set DC offset remover

        Parameters
        ----------
        dc_offset_remover : DCOffsetRemover
            A DC offset remover.

        Returns
        -------
        None
        """

        self.dc_offset_remover = dc_offset_remover

    def set_bandpass_filter(self, bandpass_filter):
        """Set bandpass filter

        Parameters
        ----------
        bandpass_filter : BandpassFilter
            A bandpass filter.

        Returns
        -------
        None
        """

        self.bandpass_filter = bandpass_filter

    def set_full_wave_rectifier(self, full_wave_rectifier):
        """Set full wave rectifier

        Parameters
        ----------
        full_wave_rectifier : FullWaveRectifier
            A full wave rectifier.

        Returns
        -------
        None
        """

        self.full_wave_rectifier = full_wave_rectifier

    def set_linear_envelope(self, linear_envelope):
        """Set linear envelope

        Parameters
        ----------
        linear_envelope : LinearEnvelope
            A linear envelope.

        Returns
        -------
        None
        """

        self.linear_envelope = linear_envelope

    def set_end_frame_cutter(self, end_frame_cutter):
        """Set end frame cutter

        Parameters
        ----------
        end_frame_cutter : EndFrameCutter
            A end frame cutter.

        Returns
        -------
        None
        """

        self.end_frame_cutter = end_frame_cutter

    def set_amplitude_normalizer(self, amplitude_normalizer):
        """Set amplitude normalizer

        Parameters
        ----------
        amplitude_normalizer : AmplitudeNormalizer
            A amplitude normalizer.

        Returns
        -------
        None
        """

        self.amplitude_normalizer = amplitude_normalizer

    def set_segmenter(self, segmenter, all_beg_ts, all_end_ts):
        """Set segmenter and beginning/end time of interest

        Parameters
        ----------
        segmenter : Segmenter
            A segmenter.

        all_beg_ts : list
            Beginning time of interest for each trial.
            See function apply_segmenter of class EMGMeasurementCollection.

        all_end_ts : list
            End time of interest for each trial.
            See function apply_segmenter of class EMGMeasurementCollection.

        Returns
        -------
        None
        """

        self.segmenter = segmenter
        self.segmenter_all_beg_ts = all_beg_ts
        self.segmenter_all_end_ts = all_end_ts

    def show_current_processes_and_related_params(self):
        """Show current processes and related parameter values

        Returns
        -------
        None
        """

        print('---- Current processes and related parameters ----')
        if self.dc_offset_remover is not None:
            print(f'DC offset remover    : {self.dc_offset_remover.get_param_values_in_str()}')
        if self.bandpass_filter is not None:
            print(f'Bandpass filter      : {self.bandpass_filter.get_param_values_in_str()}')
        if self.full_wave_rectifier is not None:
            print(f'Full wave rectifier  : {self.full_wave_rectifier.get_param_values_in_str()}')
        if self.linear_envelope is not None:
            print(f'Linear envelope      : {self.linear_envelope.get_param_values_in_str()}')
        if self.end_frame_cutter is not None:
            print(f'End frame cutter     : {self.end_frame_cutter.get_param_values_in_str()}')
        if self.amplitude_normalizer is not None:
            print(f'Amplitude normalizer : {self.amplitude_normalizer.get_param_values_in_str()}')
        if self.segmenter is not None:
            print(f'Segmenter            : {self.segmenter.get_param_values_in_str()}')

    def _plot_processing_chain(self, current_step, trial_indices_for_plot, emg_plot_params=None,
                               is_overlapping_trials=False, cycled_colors=None,
                               is_hide_legend=False, legend_kwargs=None, axes_pos_adjust=None):
        """Plot current step of the processing chain

        Parameters
        ----------
        current_step : str
            Current step in the processing chain.

        trial_indices_for_plot : list of integer
            A list of selected indices of the self.c.all_data,
            indicating which trials of the data to be plotted.

        emg_plot_params : EMGPlotParams or None, default None
            See class EMGPlotParams and function emg_plot.

        is_overlapping_trials : bool, default False
            Whether or not to plot trials of the same channel
            overlappingly on one (sub)figure.

        cycled_colors : list or None, default None
            cycled_colors is used when is_overlapping_trials is True.
            The colors for plotting overlapped trials data.

        is_hide_legend : bool, default False
            Whether or not to hide legend.

        legend_kwargs : dict or None, default None
            legend_kwargs is used when is_overlapping_trials is True.
            Parameters to control the legend display. They are the
            "other parameters" of method matplotlib.axes.Axes.legend,
            including loc, bbox_to_anchor, ncol, prop, fontsize, etc.
            (See
            https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.legend.html).

        axes_pos_adjust : 4-tuple or None, default None
            axes_pos_adjust is used when is_overlapping_trials is True.
            Parameters to adjust the axes position (i.e., plot position)
            when legend is displayed to prevent legend overlaying the plot.
            The 4-tuple represents: [0] shift of left position relative to
            width, [1] shift of bottom position relative to height, [2]
            proportion of width, [3] proportion of height.
            If None, no adjustment is applied, i.e., the value (0, 0, 1, 1)
            is applied.

        Returns
        -------
        None
        """
        if is_overlapping_trials:
            legend_labels = [self.c.trial_names[k] for k in trial_indices_for_plot]
            plot_emg_overlapping_trials(self.c.all_data, self.c.all_timestamp,
                                        trial_indices_for_plot, legend_labels,
                                        channel_names=self.c.channel_names, main_title=current_step,
                                        emg_plot_params=emg_plot_params, cycled_colors=cycled_colors,
                                        is_hide_legend=is_hide_legend, legend_kwargs=legend_kwargs,
                                        axes_pos_adjust=axes_pos_adjust)
        else:
            for k in trial_indices_for_plot:
                plot_emg(self.c.all_data[k], self.c.all_timestamp[k],
                         channel_names=self.c.channel_names,
                         main_title=f'{current_step} ({self.c.trial_names[k]})',
                         emg_plot_params=emg_plot_params)

    def process_all(self, is_plot_processing_chain=False,
                    k_for_plot=None, is_overlapping_trials=False, cycled_colors=None,
                    is_hide_legend=False, legend_kwargs=None, axes_pos_adjust=None):
        """Apply current processors to data and plot intermediate results

        Parameters
        ----------
        is_plot_processing_chain : bool
            Whether to plot the intermediate results after each
            processing step.

        k_for_plot : integer, list of integer, or None
            If integer, k_for_plot is an index of the list
            self.c.all_data.
            If list of integer, k_for_plot is a list of selected
            indices of the list self.c.all_data.
            If None, k_for_plot will be a list of all indices of the
            list self.c.all_data.
            k_for_plot sets which trials of the data to be plotted.

        is_overlapping_trials : bool, default False
            Whether or not to plot trials of the same channel
            overlappingly on one (sub)figure.

        cycled_colors : list or None, default None
            cycled_colors is used when is_overlapping_trials is True.
            The colors for plotting overlapped trials data.

        is_hide_legend : bool, default False
            Whether or not to hide legend.

        legend_kwargs : dict or None, default None
            legend_kwargs is used when is_overlapping_trials is True.
            Parameters to control the legend display. They are the
            "other parameters" of method matplotlib.axes.Axes.legend,
            including loc, bbox_to_anchor, ncol, prop, fontsize, etc.
            (See
            https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.legend.html).

        axes_pos_adjust : 4-tuple or None, default None
            axes_pos_adjust is used when is_overlapping_trials is True.
            Parameters to adjust the axes position (i.e., plot position)
            when legend is displayed to prevent legend overlaying the plot.
            The 4-tuple represents: [0] shift of left position relative to
            width, [1] shift of bottom position relative to height, [2]
            proportion of width, [3] proportion of height.
            If None, no adjustment is applied, i.e., the value (0, 0, 1, 1)
            is applied.

        Returns
        -------
        c : EMGMeasurementCollection
            An EMGMeasurementCollection instance which saves the
            processed data.
        """

        assert self.c_raw is not None, 'Data and parameters must be set by using function set_data_and_params'

        self.c = copy.deepcopy(self.c_raw)

        if k_for_plot is not None:
            trial_indices_for_plot = k_for_plot
            if not isinstance(trial_indices_for_plot, list):
                trial_indices_for_plot = [trial_indices_for_plot]  # Now this is a list
            for k in trial_indices_for_plot:
                assert k in range(len(self.c.all_data)), 'Value(s) in k_for_plot must be indices of the list all_data'
        else:
            trial_indices_for_plot = list(range(len(self.c.all_data)))  # Now this is a list of all indices all_data

        emg_plot_params = copy.deepcopy(self.c.emg_plot_params)
        if is_overlapping_trials:
            # when overlapping trials, the color in line2d_kwargs will become invalid.
            # Colors for all trials can be set by users via cycled_colors.
            if emg_plot_params.line2d_kwargs is not None:
                emg_plot_params.line2d_kwargs.pop('color', None)

        if is_plot_processing_chain:
            self._plot_processing_chain('Original', trial_indices_for_plot, emg_plot_params,
                                        is_overlapping_trials, cycled_colors,
                                        is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.dc_offset_remover is not None:
            for k in range(len(self.c.all_data)):
                self.c.all_data[k] = self.dc_offset_remover.apply(self.c.all_data[k])

            if is_plot_processing_chain:
                self._plot_processing_chain('After DC offset remover', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.bandpass_filter is not None:
            for k in range(len(self.c.all_data)):
                self.c.all_data[k] = self.bandpass_filter.apply(self.c.all_data[k])

            if is_plot_processing_chain:
                self._plot_processing_chain('After bandpass filter', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.full_wave_rectifier is not None:
            for k in range(len(self.c.all_data)):
                self.c.all_data[k] = self.full_wave_rectifier.apply(self.c.all_data[k])

            if is_plot_processing_chain:
                self._plot_processing_chain('After full wave rectifier', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.linear_envelope is not None:
            for k in range(len(self.c.all_data)):
                self.c.all_data[k] = self.linear_envelope.apply(self.c.all_data[k])

            if is_plot_processing_chain:
                self._plot_processing_chain('After linear envelope', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.end_frame_cutter is not None:
            for k in range(len(self.c.all_data)):
                self.c.all_data[k] = self.end_frame_cutter.apply(self.c.all_data[k])
                self.c.all_timestamp[k] = self.end_frame_cutter.apply(self.c.all_timestamp[k])

            if is_plot_processing_chain:
                self._plot_processing_chain('After end frame cutter', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.amplitude_normalizer is not None:
            max_amplitude = self.c.find_max_amplitude_of_each_channel_across_trials()
            for k in range(len(self.c.all_data)):
                self.c.all_data[k] = self.amplitude_normalizer.apply(self.c.all_data[k], divisor=max_amplitude)

            if is_plot_processing_chain:
                self._plot_processing_chain('After amplitude normalizer', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        if self.segmenter is not None and self.segmenter_all_beg_ts is not None \
                and self.segmenter_all_end_ts is not None:
            for k in range(len(self.c.all_data)):
                beg_idx, end_idx = BaseProcessor.get_indices_from_timestamp(
                    self.c.all_timestamp[k], self.segmenter_all_beg_ts[k], self.segmenter_all_end_ts[k])
                self.c.all_data[k] = Segmenter().apply(self.c.all_data[k], beg_idx=beg_idx, end_idx=end_idx)
                self.c.all_timestamp[k] = Segmenter().apply(self.c.all_timestamp[k], beg_idx=beg_idx, end_idx=end_idx)

            if is_plot_processing_chain:
                self._plot_processing_chain('After segmenter', trial_indices_for_plot, emg_plot_params,
                                            is_overlapping_trials, cycled_colors,
                                            is_hide_legend, legend_kwargs, axes_pos_adjust)

        return self.c



from . base import BaseProcessor


class EndFrameCutter(BaseProcessor):
    """End frame cutter for EMG signals

    Parameters
    ----------
    n_end_frames : int, default=30
        Number of frames to be cut off in both ends of the signal.
        n_end_frames >= 0.
    """

    def __init__(self, n_end_frames=30):
        assert n_end_frames >= 0, 'n_end_frames must be non-negative'
        self.n_end_frames = n_end_frames

    def apply(self, x, **kwargs):
        """Apply end frame cutter

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels),
            where n_samples > 2 * n_end_frame.
            Signal data to be processed or timestamp corresponding to
            the signal.

        Returns
        -------
        x_processed : ndarray
            Same dimension as x, where the first dimension reduces its
            length from 'n_samples' to 'n_samples - 2 * n_end_frame'.
            The result of applying end frame cutter to x.
        """

        super().assert_input(x)
        assert x.shape[0] > 2 * self.n_end_frames, 'first dimension of x must have length > 2 * n_end_frames'

        if self.n_end_frames > 0:
            x_processed = x[self.n_end_frames:(-self.n_end_frames), ]
        else:
            x_processed = x
        return x_processed

    def get_param_values_in_str(self):
        """Getting the parameter values of the processor for display
        purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = f'n_end_frames = {self.n_end_frames}'
        return params_in_str



from .. processors import *
from .. plots import *
from . emg_measurement import EMGMeasurement
import numpy as np
import copy


class EMGMeasurementCollection:
    """Wrapper of multiple-trial EMG processing

    Parameters
    ----------
    all_data : list
        Elements of all_data are signal data of the trials.
        Signal data of each trial should be ndarray of shape
        (n_samples,) or (n_samples, n_channels), where n_samples > 15
        (See Notes).
        Dimensions and n_channels (if 2-dim) of all trials should be
        the same.

    hz : float
        Sample rate in hertz.
        Ref [1] suggests that minimum sample rate be at least twice the
        highest cutoff frequency of the bandpass filter (Nyquist
        theorem) and preferably higher.
        Ref [2] suggests a minimum sample rate of 1000 Hz for surface
        EMG.
        For fine wire EMG, the authors suggest a minimum sample rate of
        2000 Hz.

    all_timestamp : list or None, default None
        If list, its length should be the same as the length of all_data
        and elements are ndarray or None.
        The ndarray in all_timestamp is the actual timestamp of the
        corresponding trial, and it should be in 1-dim and have the same
        length as the first dimension of its corresponding element in
        all_data.

    trial_names : list or None, default None
        Trial names.
        If not None, trial_names and all_data should have the same
        length.

    channel_names : list or None, default None
        If list, elements are str and its length should be equal to
        n_channels.
        Channel names of all trials to be shown in plots.

    emg_plot_params : EMGPlotParams or None, default None
        See class EMGPlotParams and function emg_plot.

    Notes
    -----
    n_samples has to meet the length condition for using bandpass
    filter. With the default parameter in BandpassFilter, n_samples
    must exceed 15. (See class BandpassFilter, function apply)

    References
    ----------
    [1] Editors (2018).
        Standards for reporting EMG data.
        Journal of Electromyography and Kinesiology, 42, I-II.
        Doi: 10.1016/S1050-6411(18)30348-1.
    [2] Stegeman, D.F., & Hermens, H.J. (1999).
        Standards for surface electromyography: the European project
        "Surface EMG for non-invasive assessment of muscles (SENIAM)".
        Biomed 2 program of the European Community. European concerted
        action. 108-112.
    """

    def __init__(self, all_data, hz, all_timestamp=None,
                 trial_names=None, channel_names=None, emg_plot_params=None):
        assert isinstance(all_data, list), 'all_data must be list'
        for k in range(len(all_data)):
            BaseProcessor.assert_input(all_data[k])
            assert all_data[k].shape[0] > 15, 'first dimension of ndarray in all_data[k] must have length > 15'
        ndims = [all_data[k].ndim for k in range(len(all_data))]
        assert all([e == ndims[0] for e in ndims]), 'Dimensions of all trials should be identical'
        if ndims[0] == 2:
            nchs = [all_data[k].shape[1] for k in range(len(all_data))]
            assert all([e == nchs[0] for e in nchs]), 'n_channels of all trials should be identical'
        self.all_data = copy.deepcopy(all_data)

        self.hz = hz

        self.all_timestamp = copy.deepcopy(all_timestamp)
        if self.all_timestamp is not None:
            assert isinstance(self.all_timestamp, list), 'all_timestamp must be list if not None'
            assert len(self.all_timestamp) == len(self.all_data), \
                'The length of all_timestamp and all_data must be identical'
        else:
            self.all_timestamp = [None] * len(self.all_data)
        for k in range(len(all_data)):
            self.all_timestamp[k] = BaseProcessor.get_timestamp(self.all_data[k].shape, self.all_timestamp[k], self.hz)

        if trial_names is None:
            self.trial_names = [None] * len(self.all_data)
        else:
            assert isinstance(trial_names, list) and len(trial_names) == len(self.all_data), \
                'trial_names should be a list with length len(all_data)'
            self.trial_names = trial_names

        self.channel_names = channel_names

        self.emg_plot_params = emg_plot_params

    def apply_dc_offset_remover(self):
        """Apply DC offset remover to the data

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            self.all_data[k] = DCOffsetRemover().apply(self.all_data[k])

    def apply_bandpass_filter(self, bf_order=4, bf_cutoff_fq_lo=10, bf_cutoff_fq_hi=450):
        """Apply bandpass filter to the data

        Parameters
        ----------
        bf_order : int, default=4
            Effective order (i.e., order after two-directional
            filtering) of the butterworth filter. bf_order should be
            a multiple of 2.

        bf_cutoff_fq_lo : float, default=10
            Low cutoff frequency of the bandpass filter.
            See class BandpassFilter.

        bf_cutoff_fq_hi : float, default=450
            High cutoff frequency of the bandpass filter.
            See class BandpassFilter.

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            self.all_data[k] = BandpassFilter(self.hz, bf_order, bf_cutoff_fq_lo, bf_cutoff_fq_hi).apply(
                self.all_data[k])

    def apply_full_wave_rectifier(self):
        """Apply full wave rectifier to the data

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            self.all_data[k] = FullWaveRectifier().apply(self.all_data[k])

    def apply_linear_envelope(self, le_order=4, le_cutoff_fq=6):
        """Apply linear envelope to the data

        Parameters
        ----------
        le_order : int, default=4
            Effective order (i.e., order after two-directional
            filtering) of the butterworth filter for linear envelope.
            le_order should be a multiple of 2.

        le_cutoff_fq : float, default=6
            Cutoff frequency of the lowpass filter.
            See class LinearEnvelope.

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            self.all_data[k] = LinearEnvelope(self.hz, le_order, le_cutoff_fq).apply(self.all_data[k])

    def apply_end_frame_cutter(self, n_end_frames=30):
        """Apply end frame cutter to the data (signal and timestamp)

        Parameters
        ----------
        n_end_frames : int, default=30
            Number of frames to be cut off in both ends of the signal.
            n_end_frames >= 0.

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            self.all_data[k] = EndFrameCutter(n_end_frames).apply(self.all_data[k])
            self.all_timestamp[k] = EndFrameCutter(n_end_frames).apply(self.all_timestamp[k])

    def find_max_amplitude_of_each_channel_across_trials(self):
        """Find max amplitude of each channel across trials

        Returns
        -------
        max_amplitude : ndarray
            Shape (n_channels,).
            Max amplitude of each channel which is found across all trials.
        """

        collect_trial_max = ()
        for k in range(len(self.all_data)):
            collect_trial_max += tuple(np.max(np.abs(self.all_data[k]), axis=0).reshape(1, -1))
        max_amplitude = np.max(np.vstack(collect_trial_max), axis=0)
        return max_amplitude

    def apply_amplitude_normalizer(self, max_amplitude):
        """Apply amplitude normalizer to the data

        Parameters
        ----------
        max_amplitude : scalar, list, or ndarray
            One or more positive values.
            If data in all_data is in 1-dim or n_channels is 1, then
            max_amplitude should be one value; otherwise max_amplitude
            should be n_channels values.
            max_amplitude is the value used as divisor in amplitude
            normalization for all trials.

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            self.all_data[k] = AmplitudeNormalizer().apply(self.all_data[k], divisor=max_amplitude)

    def apply_segmenter(self, all_beg_ts, all_end_ts):
        """Apply segmenter to the data (signal and timestamp)

        Parameters
        ----------
        all_beg_ts : list
            Elements of all_beg_ts are float and the length should be
            equal to the length of all_data.
            Beginning time of interest for each trial. For trial k,
            all_beg_ts[k] <= all_timestamp[k][-1].

        all_end_ts : list
            Elements of all_end_ts are float and the length should be
            equal to the length of all_data.
            End time of interest for each trial. For trial k,
            all_end_ts[k] >= all_timestamp[k][0] and
            all_beg_ts[k] < all_end_ts[k].

        Returns
        -------
        None
        """

        assert isinstance(all_beg_ts, list) and isinstance(all_end_ts, list), \
            'all_beg_ts and all_end_ts must be list since all_data is a list'
        assert len(all_beg_ts) == len(all_end_ts) == len(self.all_data), \
            'The length of all_beg_ts, all_end_ts, and all_data must be identical'

        for k in range(len(self.all_data)):
            beg_idx, end_idx = BaseProcessor.get_indices_from_timestamp(
                self.all_timestamp[k], all_beg_ts[k], all_end_ts[k])
            self.all_data[k] = Segmenter().apply(self.all_data[k], beg_idx=beg_idx, end_idx=end_idx)
            self.all_timestamp[k] = Segmenter().apply(self.all_timestamp[k], beg_idx=beg_idx, end_idx=end_idx)

    def __getitem__(self, k):
        """Extract data of trial k

        Parameters
        ----------
        k : index of list
            k is an integer between 0 and len(all_data) - 1.

        Returns
        -------
        m : EMGMeasurement
            An EMGMeasurement instance which possesses the data of
            trial k.
        """

        assert k in range(len(self.all_data)), 'k must be an index of all_data (a list)'

        m = EMGMeasurement(self.all_data[k], self.hz, self.all_timestamp[k],
                           self.trial_names[k], self.channel_names, self.emg_plot_params)
        return m

    def plot(self, k_for_plot=None, is_overlapping_trials=False, main_title=None,
             cycled_colors=None, is_hide_legend=False, legend_kwargs=None, axes_pos_adjust=None):
        """Plot all trials

        Parameters
        ----------
        k_for_plot : integer, list of integer, or None
            If integer, k_for_plot is an index of the list
            self.c.all_data.
            If list of integer, k_for_plot is a list of selected
            indices of the list self.c.all_data.
            If None, k_for_plot will be a list of all indices of the
            list self.c.all_data.
            k_for_plot sets which trials of the data to be plotted.

        is_overlapping_trials : bool, default False
            Whether or not to plot trials of the same channel
            overlappingly on one (sub)figure.

        main_title : str or None, default None
            main_title is used when is_overlapping_trials is True.
            The main title of the plot.

        cycled_colors : list or None, default None
            cycled_colors is used when is_overlapping_trials is True.
            The colors for plotting overlapped trials data.

        is_hide_legend : bool, default False
            Whether or not to hide legend.

        legend_kwargs : dict or None, default None
            legend_kwargs is used when is_overlapping_trials is True.
            Parameters to control the legend display. They are the
            "other parameters" of method matplotlib.axes.Axes.legend,
            including loc, bbox_to_anchor, ncol, prop, fontsize, etc.
            (See
            https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.legend.html).

        axes_pos_adjust : 4-tuple or None, default None
            axes_pos_adjust is used when is_overlapping_trials is True.
            Parameters to adjust the axes position (i.e., plot position)
            when legend is displayed to prevent legend overlaying the plot.
            The 4-tuple represents: [0] shift of left position relative to
            width, [1] shift of bottom position relative to height, [2]
            proportion of width, [3] proportion of height.
            If None, no adjustment is applied, i.e., the value (0, 0, 1, 1)
            is applied.

        Returns
        -------
        None
        """

        if k_for_plot is not None:
            trial_indices_for_plot = k_for_plot
            if not isinstance(trial_indices_for_plot, list):
                trial_indices_for_plot = [trial_indices_for_plot]  # Now this is a list
            for k in trial_indices_for_plot:
                assert k in range(len(self.all_data)), 'Value(s) in k_for_plot must be indices of the list all_data'
        else:
            trial_indices_for_plot = list(range(len(self.all_data)))  # Now this is a list of all indices all_data

        if is_overlapping_trials:
            # when overlapping trials, the color in line2d_kwargs will become invalid.
            # Colors for all trials can be set by users via cycled_colors.
            emg_plot_params = copy.deepcopy(self.emg_plot_params)
            if emg_plot_params.line2d_kwargs is not None:
                emg_plot_params.line2d_kwargs.pop('color', None)

            legend_labels = [self.trial_names[k] for k in trial_indices_for_plot]
            plot_emg_overlapping_trials(self.all_data, self.all_timestamp,
                                        trial_indices_for_plot, legend_labels,
                                        channel_names=self.channel_names, main_title=main_title,
                                        emg_plot_params=emg_plot_params, cycled_colors=cycled_colors,
                                        is_hide_legend=is_hide_legend, legend_kwargs=legend_kwargs,
                                        axes_pos_adjust=axes_pos_adjust)
        else:
            for k in trial_indices_for_plot:
                plot_emg(self.all_data[k], self.all_timestamp[k], channel_names=self.channel_names,
                         main_title=self.trial_names[k], emg_plot_params=self.emg_plot_params)

    def export_csv(self, all_csv_path):
        """Export the processing results of all trials to csv

        Parameters
        ----------
        all_csv_path : list
            The length of all_csv_path should be the same as the length
            of all_data.
            The destination paths to export data of all trials.

        Returns
        -------
        None
        """

        for k in range(len(self.all_data)):
            BaseProcessor.export_csv(all_csv_path[k], self.all_data[k],
                                     timestamp=self.all_timestamp[k], channel_names=self.channel_names)



from . base import BaseProcessor
from scipy import signal


class BandpassFilter(BaseProcessor):
    """Bandpass filter for EMG signals

    Parameters
    ----------
    hz : float
        Sample rate in hertz.
        See class EMGMeasurementCollection for suggested values of hz.

    bf_order : int, default=4
        Effective order (i.e., order after two-directional filtering)
        of the butterworth filter. bf_order should be a multiple of 2.

    bf_cutoff_fq_lo : float, default=10
        Low cutoff frequency of the bandpass filter (i.e., frequency
        higher than bf_cutoff_fq_lo will pass).
        See Notes and References for some suggestions in different
        circumstances.

    bf_cutoff_fq_hi : float, default=450
        High cutoff frequency of the bandpass filter (i.e., frequency
        lower than bf_cutoff_fq_hi will pass).
        See Notes and References for some suggestions in different
        circumstances. Note that bf_cutoff_fq_hi should be strictly
        less than half of hz.

    Notes
    -----
    Ref [1] suggests a bandwidth of 10–350 Hz for surface recording,
    10–450 Hz for intramuscular recording, and 10–1500 Hz for needle
    recording.

    Ref [2] suggests a bandwidth of from 5-10 Hz to 400-450 Hz for
    surface EMG and a high cutoff of at least 1500 Hz for
    intramuscular and needle recordings.

    Ref [3] suggests a bandwidth of from 10-20 Hz to 500-1000 Hz for
    surface EMG.

    Ref [4] suggests a low cutoff of 30 Hz when the placement of
    surface EMG is around the heart on the trunk.

    References
    ----------
    [1] Editors (2018).
        Standards for reporting EMG data.
        Journal of Electromyography and Kinesiology, 42, I-II.
        Doi: 10.1016/S1050-6411(18)30348-1.
    [2] Merletti, R, & di Torino, P. (1999).
        Standards for reporting EMG data.
        Journal of Electromyography and Kinesiology, 9(1), III-IV.
    [3] Stegeman, D.F., & Hermens, H.J. (1999).
        Standards for surface electromyography: the European project
        "Surface EMG for non-invasive assessment of muscles (SENIAM)".
        Biomed 2 program of the European Community. European concerted
        action. 108-112.
    [4] Drake, J.D.M., & Callaghan, J.P. (2006).
        Elimination of electrocardiogram contamination from
        electromyogram signals: An evaluation of currently used removal
        techniques.
        Journal of Electromyography and Kinesiology, 16, 175–187.
    """

    def __init__(self, hz, bf_order=4, bf_cutoff_fq_lo=10, bf_cutoff_fq_hi=450):
        assert hz > 2 * bf_cutoff_fq_hi, 'hz must be greater than 2 * bf_cutoff_fq_hi'
        assert bf_order % 2 == 0, 'bf_order must be a multiple of 2'

        self.hz = hz
        self.bf_order = bf_order
        self.bf_cutoff_fq_lo = bf_cutoff_fq_lo
        self.bf_cutoff_fq_hi = bf_cutoff_fq_hi

    def apply(self, x, **kwargs):
        """Apply bandpass filter

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels),
            where n_samples > (bf_order + 1) * 3 (See Notes).
            Signal data to be processed.

        Returns
        -------
        x_processed : ndarray
            Same shape as x.
            The result of applying bandpass filter to x.

        Notes
        -----
        Using scipy.signal.filtfilt requires the length of the input
        vector x greater than the padding length, padlen, which is
        (bf_order + 1) * 3 for a 'bandpass' Butterworth filter.
        """

        super().assert_input(x)
        assert x.shape[0] > (self.bf_order + 1) * 3,\
            'first dimension of x must have length > (bf_order + 1) * 3'

        b, a = signal.butter(N=int(self.bf_order/2), Wn=[self.bf_cutoff_fq_lo, self.bf_cutoff_fq_hi],
                             btype='bandpass', analog=False, output='ba', fs=self.hz)
        x_processed = signal.filtfilt(b, a, x, axis=0)  # two-directional filtering, double the order, zero phase
        return x_processed

    def get_param_values_in_str(self):
        """Getting the parameter values of the processor for display
        purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = f'hz = {self.hz}, bf_order = {self.bf_order}, bf_cutoff_fq_lo = {self.bf_cutoff_fq_lo}, ' \
                        f'bf_cutoff_fq_hi = {self.bf_cutoff_fq_hi}'
        return params_in_str



from . emg_plot import EMGPlotParams
from . emg_plot import plot_emg
from . emg_plot import plot_emg_overlapping_trials

__all__ = [
    "EMGPlotParams",
    "plot_emg",
    "plot_emg_overlapping_trials",
]



from . base import BaseProcessor
from scipy import signal


class LinearEnvelope(BaseProcessor):
    """Linear envelope for EMG signals

    Parameters
    ----------
    hz : float
        Sample rate in hertz.
        See class EMGMeasurementCollection for suggested values of hz.

    le_order : int, default=4
        Effective order (i.e., order after two-directional filtering)
        of the butterworth filter for linear envelope. le_order should
        be a multiple of 2.

    le_cutoff_fq : float, default=6
        Cutoff frequency of the lowpass filter.
        Suggested value is 1~10. For standing scenario, use 1~2.
        For running scenario, use 10.
    """

    def __init__(self, hz, le_order=4, le_cutoff_fq=6):
        assert le_order % 2 == 0, 'le_order must be a multiple of 2'

        self.hz = hz
        self.le_order = le_order
        self.le_cutoff_fq = le_cutoff_fq

    def apply(self, x, **kwargs):
        """Apply linear envelope

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels),
            where n_samples > (le_order + 1) * 3 (See Notes).
            Signal data to be processed.

        Returns
        -------
        x_processed : ndarray
            Same shape as x.
            The result of applying linear envelope to x.

        Notes
        -----
        Using scipy.signal.filtfilt requires the length of the input
        vector x greater than the padding length, padlen, which is
        (le_order / 2 + 1) * 3 for a 'lowpass' Butterworth filter.
        """

        super().assert_input(x)
        assert x.shape[0] > (self.le_order / 2 + 1) * 3,\
            'first dimension of x must have length > (le_order / 2 + 1) * 3'

        b, a = signal.butter(N=int(self.le_order/2), Wn=self.le_cutoff_fq,
                             btype='lowpass', analog=False, output='ba', fs=self.hz)
        x_processed = signal.filtfilt(b, a, x, axis=0)  # two-directional filtering, double the order, zero phase
        return x_processed

    def get_param_values_in_str(self):
        """Getting the parameter values of the processor for display
        purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = f'hz = {self.hz}, le_order = {self.le_order}, le_cutoff_fq = {self.le_cutoff_fq}'
        return params_in_str



from . test_data_processing_manager import TestDataProcessingManager
from . test_emg_measurement import TestEMGMeasurement
from . test_emg_measurement_collection import  TestEMGMeasurementCollection

__all__ = [
    'TestDataProcessingManager',
    'TestEMGMeasurement',
    'TestEMGMeasurementCollection',
]



import numpy as np
import matplotlib.pyplot as plt


class EMGPlotParams:
    """Parameters of EMG plots

    Parameters
    ----------
    n_rows : int, default None
        Number of rows to plot all channels. See function emg_plot for
        more details when considering together with data.

    n_cols : int, default None
        Number of columns to plot all channels. See function emg_plot
        for more details when considering together with data.

    fig_kwargs : dict, default None
        Parameters to control the figures. They are parameters of
        class matplotlib.figure.Figure, including figsize, dpi,
        facecolor, edgecolor, linewidth, frameon, subplotpars,
        tight_layout, constrained_layout, etc.
        (See https://matplotlib.org/stable/api/figure_api.html).

    line2d_kwargs : dict, default None
        Parameters to control the plot. They are certain parameters
        of class matplotlib.lines.Line2D, including linewidth,
        linestyle, color, marker, markersize, fillstyle, and more.
        (See
        https://matplotlib.org/stable/api/_as_gen/matplotlib.lines.Line2D.html).
    """

    def __init__(self, n_rows=None, n_cols=None, fig_kwargs=None, line2d_kwargs=None):
        self.n_rows = n_rows
        self.n_cols = n_cols
        self.fig_kwargs = fig_kwargs
        self.line2d_kwargs = line2d_kwargs


def _get_default_nrows_ncols(n_channels):
    if n_channels <= 3:
        n_cols = 1
    elif n_channels <= 10:
        n_cols = 2
    else:
        n_cols = 3
    n_rows = int(np.ceil(n_channels / n_cols))
    return n_rows, n_cols


def plot_emg(x, timestamp, channel_names=None, main_title=None, emg_plot_params=None):
    """Plot EMG signals on a created matplotlib figure

    Parameters
    ----------
    x : ndarray
        Shape (n_samples,) or (n_samples, n_channels).
        Signal data to be processed.

    timestamp : ndarray
        Shape (n_samples,).
        The timestamp corresponding to the signal.

    channel_names : list or None, default None
        If list, elements are str and its length should be equal to
        n_channels.
        Channel names to be shown in plots.

    main_title : str or None, default None
        The main title of the plot.

    emg_plot_params : EMGPlotParams or None, default None
        If this is not None, then for its n_rows and n_cols,
        (1) If n_rows and n_cols both have values, n_rows times n_cols
        should be no less than n_channels.
        (2) If exactly one of n_rows and n_cols is None, it will get
        the smallest number to fit in all channels.
        (3) If both n_rows and n_cols are None, default setting will be
        used: use one column when n_channels <= 3; use two columns
        when n_channels <= 10; otherwise use three columns.

    Returns
    -------
    None
    """

    if emg_plot_params is not None:
        n_rows, n_cols, fig_kwargs, line2d_kwargs = \
            emg_plot_params.n_rows, emg_plot_params.n_cols, emg_plot_params.fig_kwargs, emg_plot_params.line2d_kwargs
    else:
        n_rows, n_cols, fig_kwargs, line2d_kwargs = None, None, None, None

    if x.ndim == 1:
        x = x.reshape(-1, 1)
    n_channels = x.shape[1]

    if channel_names is not None:
        assert isinstance(channel_names, list) and len(channel_names) == n_channels,\
            'channel_names must be a list with length n_channels'

    if n_rows is None and n_cols is None:
        n_rows, n_cols = _get_default_nrows_ncols(n_channels)
    elif n_rows is None:
        n_rows = int(np.ceil(n_channels / n_cols))
    elif n_cols is None:
        n_cols = int(np.ceil(n_channels / n_rows))
    else:
        assert n_rows * n_cols >= n_channels, 'n_rows * n_cols >= n_channels must satisfy'

    if fig_kwargs is None:
        fig_kwargs = {}

    if line2d_kwargs is None:
        line2d_kwargs = {}

    fig, axs = plt.subplots(n_rows, n_cols, sharex='all', sharey='all', **fig_kwargs)
    plt.xlabel('Time')
    if main_title is not None:
        fig.suptitle(main_title)
    axs = np.array(axs).flat
    for i in range(n_channels):
        axs[i].plot(timestamp, x[:, i], **line2d_kwargs)
        if channel_names is not None:
            axs[i].set_title(channel_names[i])

    plt.show()


def plot_emg_overlapping_trials(all_data, all_timestamp, trial_indices_for_plot, legend_labels,
                                channel_names=None, main_title=None, emg_plot_params=None,
                                cycled_colors=None, is_hide_legend=False, legend_kwargs=None,
                                axes_pos_adjust=None):
    """Plot EMG signals on a created matplotlib figure,
    overlapping trials of the same channel

    Parameters
    ----------
    all_data : list
        Elements of all_data are signal data of the trials.
        Signal data of each trial should be ndarray of shape
        (n_samples,) or (n_samples, n_channels).
        Dimensions and n_channels (if 2-dim) of all trials should be
        the same.

    all_timestamp : list
        The length of all_timestamp should be the same as the length
        of all_data.
        Elements of all_timestamp are ndarray of shape (n_samples,).

    trial_indices_for_plot : list of integer
        A list of selected indices of all_data, indicating which trials
        of the data to be plotted.

    legend_labels : list
        The legend labels identify the trials which are overlapped in
        the plots.

    channel_names : list or None, default None
        If list, elements are str and its length should be equal to
        n_channels.
        Channel names to be shown in plots.

    main_title : str or None, default None
        The main title of the plot.

    emg_plot_params : EMGPlotParams or None, default None
        If this is not None, then for its n_rows and n_cols,
        (1) If n_rows and n_cols both have values, n_rows times n_cols
        should be no less than n_channels.
        (2) If exactly one of n_rows and n_cols is None, it will get
        the smallest number to fit in all channels.
        (3) If both n_rows and n_cols are None, default setting will be
        used: use one column when n_channels <= 3; use two columns
        when n_channels <= 10; otherwise use three columns.

    cycled_colors : list or None, default None
        The colors for plotting overlapped trials data.

    is_hide_legend : bool, default False
        Whether or not to hide legend.

    legend_kwargs : dict or None, default None
        Parameters to control the legend display. They are the
        "other parameters" of method matplotlib.axes.Axes.legend,
        including loc, bbox_to_anchor, ncol, prop, fontsize, etc.
        (See
        https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.legend.html).

    axes_pos_adjust : 4-tuple or None, default None
        Parameters to adjust the axes position (i.e., plot position)
        when legend is displayed to prevent legend overlaying the plot.
        The 4-tuple represents: [0] shift of left position relative to
        width, [1] shift of bottom position relative to height, [2]
        proportion of width, [3] proportion of height.
        If None, no adjustment is applied, i.e., the value (0, 0, 1, 1)
        is applied.

    Returns
    -------
    None
    """

    if emg_plot_params is not None:
        n_rows, n_cols, fig_kwargs, line2d_kwargs = \
            emg_plot_params.n_rows, emg_plot_params.n_cols, emg_plot_params.fig_kwargs, emg_plot_params.line2d_kwargs
    else:
        n_rows, n_cols, fig_kwargs, line2d_kwargs = None, None, None, None

    n_channels = 0
    for k in range(len(all_data)):
        if all_data[k].ndim == 1:
            all_data[k] = all_data[k].reshape(-1, 1)
        n_channels = all_data[k].shape[1]

    if channel_names is not None:
        assert isinstance(channel_names, list) and len(channel_names) == n_channels,\
            'channel_names must be a list with length n_channels'

    if n_rows is None and n_cols is None:
        n_rows, n_cols = _get_default_nrows_ncols(n_channels)
    elif n_rows is None:
        n_rows = int(np.ceil(n_channels / n_cols))
    elif n_cols is None:
        n_cols = int(np.ceil(n_channels / n_rows))
    else:
        assert n_rows * n_cols >= n_channels, 'n_rows * n_cols >= n_channels must satisfy'

    if fig_kwargs is None:
        fig_kwargs = {}

    if line2d_kwargs is None:
        line2d_kwargs = {}

    if legend_kwargs is None:
        legend_kwargs = {}

    if axes_pos_adjust is not None:
        assert isinstance(axes_pos_adjust, tuple) and len(axes_pos_adjust) == 4,\
            'axes_pos_adjust must be a tuple with length 4 if not None'
    else:
        axes_pos_adjust = (0, 0, 1, 1)

    fig, axs = plt.subplots(n_rows, n_cols, sharex='all', sharey='all', **fig_kwargs)
    plt.xlabel('Time')
    if main_title is not None:
        fig.suptitle(main_title)
    axs = np.array(axs).flat

    for i in range(n_channels):
        if cycled_colors is not None:
            axs[i].set_prop_cycle(color=cycled_colors)

        for k in trial_indices_for_plot:
            axs[i].plot(all_timestamp[k], all_data[k][:, i], **line2d_kwargs)

        box = axs[i].get_position()
        axs[i].set_position([box.x0 + box.width * axes_pos_adjust[0],
                             box.y0 + box.height * axes_pos_adjust[1],
                             box.width * axes_pos_adjust[2],
                             box.height * axes_pos_adjust[3]])
        if not is_hide_legend:
            axs[i].legend(labels=legend_labels, **legend_kwargs)

        if channel_names is not None:
            axs[i].set_title(channel_names[i])

    plt.show()



from . data_processing_manager import DataProcessingManager
from . emg_measurement import EMGMeasurement
from . emg_measurement_collection import EMGMeasurementCollection

__all__ = [
    "DataProcessingManager",
    "EMGMeasurement",
    "EMGMeasurementCollection",
]



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='pyemgpipeline',
    version='1.0.0',
    author='tlwu et al',
    author_email='tlwu2008@gmail.com',
    description='EMG signal processing pipeline',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/aalhossary/pyemgpipeline',
    project_urls={
        'API reference': 'https://aalhossary.github.io/pyemgpipeline/api.html',
        'Bug Tracker': 'https://github.com/aalhossary/pyemgpipeline/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(),
    install_requires=['numpy', 'scipy', 'matplotlib'],
    python_requires='>=3.6',
)



from . base import BaseProcessor
import numpy as np


class Segmenter(BaseProcessor):
    """Segmenter for EMG signals

    """

    def __init__(self):
        pass

    def apply(self, x, **kwargs):
        """Apply segmenter

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels).
            Signal data to be processed or timestamp corresponding to
            the signal.

        beg_idx : int
            Beginning index (inclusive) for segmentation of x at its
            first dimension. 0 <= beg_idx <= end_idx < n_samples.

        end_idx : int
            End index (inclusive) for segmentation of x at its first
            dimension. 0 <= beg_idx <= end_idx < n_samples.

        Returns
        -------
        x_processed : ndarray
            Same dimension as x, where the first dimension reduces its
            length from 'n_samples' to 'end_idx - beg_idx + 1'.
            The result of applying segmenter to x.
        """

        super().assert_input(x)

        assert 'beg_idx' in kwargs.keys(), 'Need to provide parameter "beg_idx" in Segmenter.apply'
        assert 'end_idx' in kwargs.keys(), 'Need to provide parameter "end_idx" in Segmenter.apply'
        beg_idx = kwargs['beg_idx']
        end_idx = kwargs['end_idx']
        assert isinstance(beg_idx, (int, np.integer)) and isinstance(end_idx, (int, np.integer))\
               and 0 <= beg_idx <= end_idx < x.shape[0],\
               'beg_idx and end_idx must be integers and 0 <= beg_idx <= end_idx < x.shape[0]'

        x_processed = x[beg_idx:(end_idx + 1), ]
        return x_processed

    def get_param_values_in_str(self):
        """Getting the parameter values of the processor for display
        purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = super().get_param_values_in_str()
        return params_in_str



from unittest import TestCase
import numpy as np
from pyemgpipeline.wrappers import DataProcessingManager
from pyemgpipeline.processors import BandpassFilter, LinearEnvelope, EndFrameCutter
from pyemgpipeline.processors import AmplitudeNormalizer, Segmenter


class TestDataProcessingManager(TestCase):
    def setUp(self):
        all_data = [
            np.array([20.3, 41.0, 53.9, 63.3, 39.5, 24.9, 26.1, 24.0, 44.1, 42.0,
                      37.4, 24.6, -21.8, -56.3, -48.1, -45.0, -29.1, -9.6, 5.3, 1.4]),
            np.array([1.1, 2.9, 1.4, -0.2, 4.4, 7.2, 9.9, 19.1, 14.2, 18.8,
                      17.2, 17.9, 11.1, 13.9, 15.4, 19.4, 12.1, 16.9, 12.4, 9.0])
        ]
        hz = 1000
        self.mgr = DataProcessingManager()
        self.mgr.set_data_and_params(all_data, hz=hz)
        self.mgr.set_bandpass_filter(BandpassFilter(hz, bf_order=4, bf_cutoff_fq_lo=10, bf_cutoff_fq_hi=450))
        self.mgr.set_linear_envelope(LinearEnvelope(hz, le_order=4, le_cutoff_fq=5))
        self.mgr.set_end_frame_cutter(EndFrameCutter(n_end_frames=3))
        self.mgr.set_amplitude_normalizer(AmplitudeNormalizer())
        all_beg_ts = [0.006, 0.0055]
        all_end_ts = [0.012, 0.0095]
        self.mgr.set_segmenter(Segmenter(), all_beg_ts, all_end_ts)

    def tearDown(self):
        pass

    def test_process_all(self):
        c = self.mgr.process_all(is_plot_processing_chain=False)
        actual = c.all_data
        desired = [
            np.array([0.9705620, 0.9753172, 0.9796631, 0.9835977, 0.9871213, 0.9902362, 0.9929473]),
            np.array([-0.6566767, -0.6513617, -0.6464390, -0.6419096])
        ]
        self.assertEqual(len(actual), len(desired))
        for i in range(len(actual)):
            np.testing.assert_allclose(actual[i], desired[i])



from abc import ABCMeta, abstractmethod
import numpy as np


class BaseProcessor(metaclass=ABCMeta):
    """Base class of all EMG signal processors in pyemgpipeline
    """

    @abstractmethod
    def apply(self, x, **kwargs):
        """Abstract method of applying any EMG signal processing steps

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels),
            where processors may require conditions on n_samples.
            Signal data to be processed.

        **kwargs : keyword arguments
            Processor-specific arguments.

        Returns
        -------
        None (in abstract method)
        """

        return

    @abstractmethod
    def get_param_values_in_str(self):
        """Abstract method of getting the parameter values of the
        processor for display purpose

        Returns
        -------
        params_in_str : str
            Parameter values.
        """

        params_in_str = 'No parameters'
        return params_in_str

    @staticmethod
    def assert_input(x):
        """Check the input signal

        Parameters
        ----------
        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels).
            Signal data to be processed.

        Returns
        -------
        None
        """

        assert isinstance(x, np.ndarray), 'x must be an ndarray'
        assert x.ndim == 1 or x.ndim == 2, 'x must be a 1-dim or 2-dim ndarray'
        assert not np.isnan(np.sum(x)), 'x must not contain nan values'

    @staticmethod
    def get_timestamp(x_shape, timestamp=None, hz=None):
        """Get the timestamp from one of the arguments

        Parameters
        ----------
        x_shape : tuple
            Shape of the signal data in ndarray.
            x_shape is (n_samples,) or (n_samples, n_channels).

        timestamp : ndarray or None, default None
            The actual timestamp corresponding to the signal. If it is
            an ndarray, it should be in 1-dim and have length equal to
            x_shape[0].

        hz : float or None, default None
            Sample rate in hertz.

        Returns
        -------
        timestamp: ndarray
            Shape (n_samples,).
            The deduced timestamp corresponding to the signal. If
            getting None for both timestamp and hz from the input,
            this return value will be integers starting from 0.
        """

        if timestamp is not None:
            assert isinstance(timestamp, np.ndarray), 'timestamp must be an ndarray'
            assert timestamp.ndim == 1 and timestamp.shape[0] == x_shape[0], \
                'timestamp must be in 1-dim and have length equal to x_shape[0]'
            assert not np.isnan(np.sum(timestamp)), 'timestamp must not contain nan values'
        elif hz is not None:
            assert isinstance(hz, float) or isinstance(hz, int), 'hz must be a number'
            assert hz > 0, 'hz must be positive'
            timestamp = np.array([e / hz for e in range(x_shape[0])])
        else:
            timestamp = np.array(range(x_shape[0]))
        return timestamp

    @staticmethod
    def get_indices_from_timestamp(timestamp, beg_ts, end_ts):
        """Get the corresponding indices from timestamp

        Parameters
        ----------
        timestamp : ndarray
            In 1-dim.

        beg_ts : float
            Beginning time of interest. beg_ts <= timestamp[-1].

        end_ts : float
            End time of interest. end_ts >= timestamp[0].
            Also beg_ts < end_ts.

        Returns
        -------
        beg_idx : int
            The smallest index of timestamp in which the time is
            greater than or equal to beg_ts.

        end_idx : int
            The largest index of timestamp in which the time is
            less than or equal to end_ts.
        """

        assert beg_ts < end_ts, 'beg_ts must be less than end_ts'
        assert beg_ts <= timestamp[-1] and end_ts >= timestamp[0],\
            'beg_ts must be no greater than timestamp[-1] and end_ts must be no less than timestamp[0]'

        beg_idx = int(np.where(timestamp >= beg_ts)[0][0])
        end_idx = int(np.where(timestamp <= end_ts)[0][-1])
        return beg_idx, end_idx

    @staticmethod
    def export_csv(filepath, x, timestamp=None, channel_names=None):
        """Export data to csv file

        Parameters
        ----------
        filepath : str
            Filepath for data to export.

        x : ndarray
            Shape (n_samples,) or (n_samples, n_channels).
            Signal data to be exported.

        timestamp : ndarray or None, default None
            If ndarray, shape is (n_samples,).
            The timestamp corresponding to the signal.

        channel_names : list of str, or None, default None
            If list, its length should be equal to n_channels.
            Channel names to be shown as the header of the csv file.

        Returns
        -------
        None
        """

        if x.ndim == 1:
            x = x.reshape(-1, 1)
        n_channels = x.shape[1]

        if timestamp is not None:
            out_data = np.hstack((timestamp.reshape(-1, 1), x))
        else:
            out_data = x

        if channel_names is not None:
            assert isinstance(channel_names, list) and len(channel_names) == n_channels, \
                'channel_names must be a list with length n_channels'
            header = ','.join(channel_names)
            if timestamp is not None:
                header = 'Timestamp,' + header
        else:
            header = ''

        np.savetxt(filepath, out_data, delimiter=',', header=header, comments='')
